from ftw.geo.setuphandlers import uninstall_products_maps
from ftw.upgrade import UpgradeStep


class UninstallProductsMaps(UpgradeStep):
    """Uninstalls Products.Maps using the quickinstaller tool.
    """

    def __call__(self):
        uninstall_products_maps(self.portal)



from mocker import ARGS
from mocker import Expect
from mocker import KWARGS
from mocker import Mocker


class ExpectGeocodingRequest(object):
    """Mock geopy requests for geocoding a location.

    Example:

    with ExpectGeocodingRequest('Bern, Switzerland', (46.9479222, 7.444608499999999)):
        do_something()
    """

    def __init__(self, place='Bern, Switzerland', coords=(46.947922, 7.444608)):
        self.place = place
        self.coords = coords

    def __enter__(self):
        self.mocker = Mocker()
        expect = Expect(self.mocker)
        method = self.mocker.replace('ftw.geo.handlers.geocode_location')
        expect(method(ARGS, KWARGS)).result((self.place, self.coords, None))
        self.mocker.replay()

    def __exit__(self, exc_type, exc_value, traceback):
        if not exc_type:
            self.mocker.verify()
            self.mocker.restore()



from Products.ATContentTypes.content.document import ATDocument
from Products.ATContentTypes.interfaces import IATDocument
from collective.geo.contentlocations.interfaces import IGeoManager
from collective.geo.geographer.interfaces import IGeoreferenceable
from ftw.geo.interfaces import IGeocodableLocation
from ftw.geo.testing import GEO_FUNCTIONAL_TESTING
from ftw.geo.tests.helpers import ExpectGeocodingRequest
from ftw.testbrowser import browsing
from ftw.testbrowser.pages import factoriesmenu
from ftw.testbrowser.pages import statusmessages
from plone.app.testing import TEST_USER_ID
from plone.app.testing import TEST_USER_NAME
from plone.app.testing import login
from plone.app.testing import setRoles
from unittest2 import TestCase
from zope.component import getSiteManager
from zope.interface import classImplements
import transaction


class TitleLocationAdapter(object):
    """Use the title for geocoding in the test.
    """

    def __init__(self, context):
        self.context = context

    def getLocationString(self):
        return self.context.Title()


class TestArchetypesEvents(TestCase):

    layer = GEO_FUNCTIONAL_TESTING

    def setUp(self):
        self.portal = self.layer['portal']
        setRoles(self.portal, TEST_USER_ID, ['Manager'])
        login(self.portal, TEST_USER_NAME)

        getSiteManager().registerAdapter(factory=TitleLocationAdapter,
                                         required=[IATDocument],
                                         provided=IGeocodableLocation)
        classImplements(ATDocument, IGeoreferenceable)
        transaction.commit()

    def tearDown(self):
        getSiteManager().unregisterAdapter(factory=TitleLocationAdapter,
                                           required=[IATDocument],
                                           provided=IGeocodableLocation)
        transaction.commit()

    @browsing
    def test_geocoding_triggered_when_creating_object(self, browser):
        browser.login().open()
        factoriesmenu.add('Page')
        with ExpectGeocodingRequest():
            browser.fill({'Title': 'Bern, Switzerland'}).submit()
        statusmessages.assert_no_error_messages()

        obj = self.portal.get('bern-switzerland')
        self.assertEquals(('Point', (7.444608, 46.947922)),
                          IGeoManager(obj).getCoordinates())

    @browsing
    def test_geocoding_triggered_when_editing_address(self, browser):
        browser.login().open()
        factoriesmenu.add('Page')
        with ExpectGeocodingRequest():
            browser.fill({'Title': 'Bern, Switzerland'}).submit()
        statusmessages.assert_no_error_messages()
        obj = self.portal.get('bern-switzerland')

        browser.find('Edit').click()
        with ExpectGeocodingRequest('Zurich, Switzerland', (8.53918, 47.36864)):
            browser.fill({'Title': 'Zurich, Switzerland'}).submit()

        self.assertEquals(('Point', (47.36864, 8.53918)),
                          IGeoManager(obj).getCoordinates())

    @browsing
    def test_delete_geocoding_if_address_removed(self, browser):
        browser.login().open()
        factoriesmenu.add('Page')

        with ExpectGeocodingRequest():
            browser.fill({'Title': 'Bern, Switzerland'}).submit()

        statusmessages.assert_no_error_messages()
        obj = self.portal.get('bern-switzerland')
        obj.schema.get('title').required = False

        browser.find('Edit').click()
        browser.fill({'Title': ''}).submit()

        self.assertEquals((None, None), IGeoManager(obj).getCoordinates())



# pylint: disable=E0211
# E0211: Method has no argument

from zope.interface import Interface
from zope.viewlet.interfaces import IViewletManager


class IGeocodableLocation(Interface):
    """Interface used to implement an adapter that is able to represent a
    content item's location as a geocodable string.

    In order for content types to be automatically geocoded on ``ObjectEdited``
    or ``ObjectInitialized`` events, there must be an adapter for the content
    type that implements ``IGeocodableLocation`` and knows how to build a
    geocodable location string from the content type's location related fields.

    In order to implement the interface the adapter needs to define a
    ``getLocationString`` method that returns the complete location as a comma
    separated string, with the location parts getting less specific from
    left to right.

    For example:

        '1600 Amphitheatre Parkway, Mountain View, CA, US'
        'Engehaldestr. 53, 3012 Bern, Switzerland'

    If the ``getLocationString`` method returns the empty string or ``None``,
    the event handler won't attempt to do a geocode lookup, so this is the
    suggested way to abort geocoding if not enough location information is
    available.
    """

    def getLocationString():
        """Build a geocodable location string from the content type's location
        related fields.
        """


class IContentMapViewletManager(IViewletManager):
    """A viewlet manager that renders a map for a Georeferenceable content.
    """



from setuptools import setup, find_packages
import os

version = '1.3.6'
maintainer = 'Lukas Graf'

tests_require = [
    'ftw.testbrowser',
    'ftw.testing',
    'mocker',
    'plone.app.testing',
    'plone.testing',
    'zope.configuration',
    'unittest2',
    'transaction',
    'plone.app.referenceablebehavior',
    'plone.directives.form',
    ]

setup(name='ftw.geo',
      version=version,
      description='Integration package for collective.geo.* packages.',
      long_description=open('README.rst').read() + '\n' + \
          open(os.path.join('docs', 'HISTORY.txt')).read(),

      classifiers=[
        'Framework :: Plone',
        'Framework :: Plone :: 4.2',
        'Framework :: Plone :: 4.3',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License (GPL)',
        'Programming Language :: Python',
        'Topic :: Software Development',
        ],

      keywords='plone ftw geo',
      author='4teamwork AG',
      author_email='mailto:info@4teamwork.ch',
      maintainer=maintainer,
      url='https://github.com/4teamwork/ftw.geo',
      license='GPL2',

      packages=find_packages(exclude=['ez_setup']),
      namespace_packages=['ftw'],
      include_package_data=True,
      zip_safe=False,

      install_requires=[
        'setuptools',
        'geopy',
        'collective.geo.settings',
        'collective.geo.openlayers >= 3.0, < 4.0',
        'collective.geo.geographer',
        'collective.geo.contentlocations',
        'collective.geo.kml',
        'collective.geo.mapwidget >= 2.1, < 3.0',
        'plone.app.dexterity',
        'plone.api',
        'ftw.upgrade',
        ],

      tests_require=tests_require,
      extras_require=dict(tests=tests_require),

      entry_points='''
      # -*- Entry points: -*-
      [z3c.autoinclude.plugin]
      target = plone
      ''',
      )



from collective.geo.contentlocations.interfaces import IGeoManager
from collective.geo.geographer.interfaces import IGeoreferenceable
from collective.geo.geographer.interfaces import IGeoreferenced
from collective.geo.settings.interfaces import IGeoSettings
from ftw.geo.handlers import geocodeAddressHandler
from ftw.geo.interfaces import IGeocodableLocation
from ftw.geo.testing import ZCML_LAYER
from ftw.geo.tests.utils import is_coord_tuple
from ftw.testing import MockTestCase
from mocker import ANY
from mocker import ARGS
from mocker import KWARGS
from mocker import MATCH
from plone import api
from plone.memoize import ram
from plone.registry.interfaces import IRegistry
from Products.statusmessages.interfaces import IStatusMessage
from urllib2 import URLError
from ZODB.POSException import ConflictError
from zope.annotation.interfaces import IAnnotations
from zope.component import adapts
from zope.component import getGlobalSiteManager
from zope.component import provideAdapter
from zope.component import queryAdapter
from zope.component.hooks import setSite
from zope.interface import implements
from zope.interface import Interface
from zope.interface.verify import verifyClass


try:
    # geopy < 0.96
    from geopy.geocoders.googlev3 import GQueryError
    from geopy.geocoders.googlev3 import GTooManyQueriesError
    GeocoderQueryError = GQueryError
    GeocoderQuotaExceeded = GTooManyQueriesError
except ImportError:
    # geopy >= 0.96
    from geopy.exc import GeocoderQueryError
    from geopy.exc import GeocoderQuotaExceeded


class ISomeType(Interface):
    pass


class SomeTypeLocationAdapter(object):
    """Adapter that is able to represent the location of a SomeType in
    a geocodable string form.
    """
    implements(IGeocodableLocation)
    adapts(ISomeType)

    def __init__(self, context):
        self.context = context

    def getLocationString(self):
        """Build a geocodable location string from SomeType's address
        related fields.
        """
        street = self.context.getAddress()
        zip_code = self.context.getZip()
        city = self.context.getCity()
        country = self.context.getCountry()

        if not (street or zip_code or city):
            # Not enough location information to do sensible geocoding
            return ''

        location = ', '.join([street, zip_code, city, country])
        return location


class TestGeocoding(MockTestCase):

    layer = ZCML_LAYER

    def setUp(self):
        super(TestGeocoding, self).setUp()
        provideAdapter(SomeTypeLocationAdapter)
        self.context = None

        self.original_get_registry_record = api.portal.get_registry_record
        api.portal.get_registry_record = lambda *args, **kwargs: 'mock tests deserve this'

    def tearDown(self):
        super(TestGeocoding, self).tearDown()
        self.context = None
        # Invalidate the geocoding cache so tests run in isolation
        ram.global_cache.invalidate('ftw.geo.handlers.geocode_location')

        api.portal.get_registry_record = self.original_get_registry_record

    def replace_geopy_geocoders(self, result=None):
        """Replace the geocode method of the Google geocoder with a mock
        that doesn't actually send a request to the Google API.
        """

        if not result:
            # Use a default result
            result = ((u'3012 Berne, Switzerland',
                      (46.958857500000001, 7.4273286000000001)), )

        self.request = self.mocker.mock()
        req_method = self.mocker.replace(
            'geopy.geocoders.googlev3.GoogleV3.geocode')
        self.expect(req_method(ARGS, KWARGS)).call(
            self.request).count(0, None)
        self.expect(self.request(ARGS, KWARGS)).result(result)

    def mock_site(self):
        site = self.create_dummy(getSiteManager=getGlobalSiteManager,
                                 REQUEST=self.stub_request())
        setSite(site)

    def mock_statusmessage_adapter(self):
        self.statusmsg = self.mocker.mock(count=False)
        self.message_cache = self.create_dummy()
        self.expect(self.statusmsg(ANY).addStatusMessage(ANY, type=ANY)).call(
            lambda msg, type: setattr(self.message_cache, type, msg))
        self.mock_adapter(self.statusmsg, IStatusMessage, (Interface, ))

    def mock_context(self, address='Engehaldestr. 53',
                     zip_code='3012',
                     city='Bern',
                     country='Switzerland'):
        ifaces = [ISomeType, IGeoreferenceable, IAnnotations, IGeoreferenced]
        self.context = self.providing_stub(ifaces)
        self.expect(self.context.getAddress()).result(address)
        self.expect(self.context.getZip()).result(zip_code)
        self.expect(self.context.getCity()).result(city)
        self.expect(self.context.getCountry()).result(country)

        request = self.stub_request()
        self.expect(self.context.REQUEST).result(request)

    def mock_annotations(self, count=1):
        annotation_factory = self.mocker.mock()
        self.mock_adapter(annotation_factory, IAnnotations, (Interface,))
        self.expect(annotation_factory(self.context)).result({}).count(count)

    def mock_geosettings_registry(self, api_key=None):
        registry = self.stub()
        self.mock_utility(registry, IRegistry)
        proxy = self.stub()
        self.expect(registry.forInterface(IGeoSettings)).result(proxy)
        self.expect(proxy.googleapi).result(api_key)

    def mock_geomanager(self, count=1):
        geomanager_proxy = self.stub()
        geomanager_factory = self.mocker.mock()
        self.mock_adapter(geomanager_factory, IGeoManager, (Interface,))
        self.expect(geomanager_factory(self.context)
                    ).result(geomanager_proxy).count(0, None)
        coords = ('Point', MATCH(is_coord_tuple))
        self.expect(geomanager_proxy.setCoordinates(*coords)).count(count)

    def test_geocoding_adapter(self):
        self.mock_context()
        self.replay()

        location_adapter = queryAdapter(self.context, IGeocodableLocation)
        self.assertTrue(location_adapter is not None)

        loc_string = location_adapter.getLocationString()
        self.assertEquals(loc_string,
                          'Engehaldestr. 53, 3012, Bern, Switzerland')

        verifyClass(IGeocodableLocation, SomeTypeLocationAdapter)

    def test_geocoding_handler(self):
        self.mock_site()
        self.mock_context()
        self.mock_geomanager()
        self.mock_annotations()
        self.mock_geosettings_registry()
        self.replace_geopy_geocoders()
        self.replay()

        event = self.mocker.mock()
        geocodeAddressHandler(self.context, event)

    def test_geocoding_handler_with_same_location(self):
        self.mock_site()
        # Use different address values for context to avoid caching
        self.mock_context('Hirschengraben', '3000', 'Bern', 'Switzerland')
        # geo manager should only be called once since the second request
        # won't cause a lookup
        self.mock_geomanager(count=1)
        self.mock_annotations(count=2)
        self.mock_geosettings_registry()
        self.replace_geopy_geocoders()
        event = self.mocker.mock()
        self.replay()

        # Call the handler twice with the same context, shouldn't cause a
        # lookup since location didn't change.
        geocodeAddressHandler(self.context, event)
        geocodeAddressHandler(self.context, event)

    def test_geocoding_handler_with_api_key(self):
        self.mock_site()
        # Use different address values for context to avoid caching
        self.mock_context('Bahnhofplatz', '3000', 'Bern', 'Switzerland')
        self.mock_geomanager()
        self.mock_annotations()
        self.mock_geosettings_registry(api_key='API_KEY_123')
        self.replace_geopy_geocoders()
        event = self.mocker.mock()
        self.replay()

        geocodeAddressHandler(self.context, event)

    def test_geocoding_handler_with_invalid_location(self):
        self.mock_site()
        self.mock_statusmessage_adapter()
        self.mock_context('Bag End', '1234', 'The Shire', 'Middle Earth')
        self.mock_annotations()
        self.mock_geosettings_registry()
        self.replace_geopy_geocoders()
        self.mocker.throw(GeocoderQueryError)
        event = self.mocker.mock()
        self.replay()

        geocodeAddressHandler(self.context, event)
        # Expect the appropriate info message
        self.assertEquals(self.message_cache.info, 'msg_no_match')

    def test_geocoding_handler_with_invalid_non_ascii_locatio(self):
        self.mock_site()
        self.mock_statusmessage_adapter()
        self.mock_context('Ober\xc3\xa4geri', '1234', 'The Shire', 'Middle Earth')
        self.mock_annotations()
        self.replace_geopy_geocoders()
        self.mocker.throw(GeocoderQueryError)
        event = self.mocker.mock()
        self.replay()

        geocodeAddressHandler(self.context, event)
        # Expect the appropriate info message
        self.assertEquals(self.message_cache.info, 'msg_no_match')

        msg = self.message_cache.info
        loc = msg.mapping['location']
        # Concatenate message (unicode) and location to force a possible
        # UnicodeDecodeError if string types don't match
        self.assertIsInstance(msg + loc, unicode)

    def test_geocoding_handler_with_empty_location_string(self):
        self.mock_site()
        self.mock_context('', '', '', '')
        self.mock_geosettings_registry()
        event = self.mocker.mock()
        self.replay()

        geocodeAddressHandler(self.context, event)

    def test_geocoding_handler_with_missing_adapter(self):
        self.mock_site()
        self.mock_context()
        # Unregister the IGeocodableLocation adapter
        gsm = getGlobalSiteManager()
        gsm.unregisterAdapter(SomeTypeLocationAdapter)
        event = self.mocker.mock()
        self.replay()

        # Handler should not fail even though there is no adapter
        geocodeAddressHandler(self.context, event)

    def test_geocoding_handler_with_too_many_queries(self):
        self.mock_site()
        self.mock_statusmessage_adapter()
        self.mock_context('Some Address')
        self.mock_annotations()
        self.mock_geosettings_registry()
        self.replace_geopy_geocoders()

        self.mocker.throw(GeocoderQuotaExceeded)
        event = self.mocker.mock()
        self.replay()

        geocodeAddressHandler(self.context, event)
        # Expect the appropriate info message
        self.assertEquals(self.message_cache.info, 'msg_too_many_queries')

    def test_multiple_results(self):
        self.mock_site()
        self.mock_statusmessage_adapter()
        self.mock_context('Hasslerstrasse', '3000', 'Bern', 'Switzerland')
        self.mock_geomanager()
        self.mock_annotations()
        self.mock_geosettings_registry()

        result = ((u'3001 Berne, Switzerland',
                   (46.958857500000001, 7.4273286000000001)),
                  (u'3000 Berne, Switzerland',
                   (46.958857500000002, 7.4273286000000002)), )
        self.replace_geopy_geocoders(result=result)
        self.replay()

        geocodeAddressHandler(self.context, None)
        # Expect the appropriate info message
        self.assertEquals(self.message_cache.info, 'msg_multiple_matches')

    def test_geocoding_handler_with_network_error(self):
        self.mock_site()
        self.mock_statusmessage_adapter()
        self.mock_context('Some Address')
        self.mock_annotations()
        self.mock_geosettings_registry()
        self.replace_geopy_geocoders()

        event = self.mocker.mock()
        self.mocker.throw(URLError('foo'))
        self.replay()

        geocodeAddressHandler(self.context, event)
        # Expect the appropriate info message
        self.assertEquals(self.message_cache.info, 'msg_network_error')

    def test_geocoding_doesnt_swallow_conflict_error(self):
        self.mock_site()
        self.mock_context('Some Address')
        self.mock_annotations()
        self.mock_geosettings_registry()
        self.replace_geopy_geocoders()

        self.mocker.throw(ConflictError)
        event = self.mocker.mock()
        self.replay()

        # Make sure ConflictError always gets raised
        with self.assertRaises(ConflictError):
            geocodeAddressHandler(self.context, event)

    def test_geocoding_unhandled_exception(self):
        self.mock_site()
        self.mock_statusmessage_adapter()
        self.mock_context('Some Address')
        self.mock_annotations()
        self.mock_geosettings_registry()
        self.replace_geopy_geocoders()

        event = self.mocker.mock()
        self.mocker.throw(Exception('Something broke!'))
        self.replay()

        geocodeAddressHandler(self.context, event)
        # Expect the appropriate info message
        self.assertEquals(self.message_cache.info, 'msg_unhandled_exception')



"""
Various utils used in ftw.geo tests
"""


def is_coord_tuple(value):
    """Check if `value` is a coordinate tuple, like for example
    (7.4273286000000001, 46.958857500000001)
    """
    if isinstance(value, tuple) and len(value) == 2:
        if isinstance(value[0], float) and \
           isinstance(value[1], float):
            return True
    return False